rawset(_G, "CONFIG_Queue", CONFIG_Queue or {})

local function CONFIG_RegisterVar(var)
	table.insert(CONFIG_Queue, var)
	return CV_RegisterVar(var)
end

-- Sprays by GenericHeroGuy

freeslot("MT_SPRAY", "SPR_GRFT", "SPR_GRFU", "sfx_sprych", "sfx_sprygt")

mobjinfo[MT_SPRAY] = {
	flags = MF_NOBLOCKMAP|MF_NOGRAVITY|MF_NOCLIPHEIGHT|MF_DONTENCOREMAP,
	spawnstate = S_INVISIBLE,
	radius = 8*FRACUNIT,
	height = 128*FRACUNIT,
	dispoffset = 4 -- would've been 2 if it wasn't for OpenGL
}

local CHARGE_TIME = TICRATE * 2 -- should this be a cvar? i dunno
local BT_SPRAY = 1<<12 -- we claimin' unused bits now?????
local starttime = 6*TICRATE + (3*TICRATE/4)
local detectsprites = true
local showtip = true
local tiptimer = 0
local framecount = 0
local detectedframes = { "GRFTA" } -- for dedis
local usespray = {}
local visibilitychanged = false
local stickychanged = false

-- sounds
local LAUNCHSOUND = sfx_s266 -- spray gets launched
local SPRAYEDSOUND = sfx_s22e -- something (or someone) gets sprayed
local CHARGESOUND = sfx_sprych -- sprays charging up
local GOTSOUND = sfx_sprygt -- got sprays
local NOSPAWNSOUND = sfx_s22b -- spray can't spawn
local DESTROYSOUND = sfx_s22c -- spray disintegrates after being stuck on a player

-- the other part of P_RandomRange
-- dedis don't have HUD hooks of course, so we need this to recalculate the spray's frame
-- from the RNG value we got (in movecount)
local function NotRandomRange(n, a, b)
	return (n * (b-a+1)) >> FRACBITS + a
end

-- enable locally
local cv_sprayenabled = CONFIG_RegisterVar{
	name = "showsprays",
	defaultvalue = "On",
	flags = CV_CALL|CV_NOINIT,
	possiblevalue = CV_OnOff,
	func = function()
		visibilitychanged = true
		stickychanged = true
	end,

	config_menu = "Sprays",
	displayname = "Show sprays",
	description = "Show or hide sprays."
}

-- show a tip for using sprays
local cv_spraytip = CONFIG_RegisterVar{
	name = "spraytip",
	defaultvalue = "On",
	possiblevalue = CV_OnOff,

	config_menu = "Sprays",
	displayname = "Show tip",
	description = "Show a tip when you get sprays for the first time in a session."
}

-- button to press
local cv_spraybutton = {}
local possible = {
	None = 0,
	["Custom 1"] = BT_CUSTOM1,
	["Custom 2"] = BT_CUSTOM2,
	["Custom 3"] = BT_CUSTOM3,
	["Custom 1+2"] = BT_CUSTOM1|BT_CUSTOM2,
	["Custom 2+3"] = BT_CUSTOM2|BT_CUSTOM3,
	["Custom 1+3"] = BT_CUSTOM1|BT_CUSTOM3,
	Item = BT_ATTACK,
}
cv_spraybutton[0] = CONFIG_RegisterVar{
	name = "spraybutton",
	defaultvalue = "Custom 2",
	possiblevalue = possible,

	config_menu = "Sprays",
	displayname = "Button",
	description = "Which button to use for firing sprays. (alternatively, use console command \"usespray\")",
	config_multi = { "spraybutton2", "spraybutton3", "spraybutton4" }
}
for i = 2, 4 do
	cv_spraybutton[i-1] = CV_RegisterVar({
		name = "spraybutton"..i,
		defaultvalue = "Custom 2",
		possiblevalue = possible,
	})
end

-- HUD offsets for spray can
local cv_sprayoffsetx = CONFIG_RegisterVar{
	name = "sprayoffset_x",
	defaultvalue = 0,

	config_menu = "Sprays",
	displayname = "HUD offset",
	description = "Change offsets for the spray can.",
	config_hudmove = "sprayoffset_y",
	config_hudcall = function(v, p, c, xvar, yvar)
		for dp, splitindex in displayplayers.iterate do
			local vflags = SG_GetSplitscreenFlags(splitindex) | V_HUDTRANS | V_SNAPTOLEFT | V_SNAPTOTOP
			local x, y = 42, 64
			local align = "center"
			local colormap = v.getColormap(TC_DEFAULT, dp.skincolor)

			if splitscreen == 1 then
				x, y = 68, 15
			elseif splitscreen >= 2 then
				-- 3P/4P splitscreen is text only
				vflags = $ | V_GREENMAP
				x, y = 30, 4
				align = "left"
				if splitindex & 1 == 1 then
					-- flip to other side
					vflags = $ | V_SNAPTOLEFT ^^ V_SNAPTORIGHT
					x = 160 - $
					align = "right"
				end
			end

			x, y = x + xvar.value, y + yvar.value

			if splitscreen <= 1 then v.draw(x - 12, y - 8, v.cachePatch("SPRAYCAN"), vflags, colormap) end
			v.drawString(x, y, "123", vflags, align)
		end
	end
}
local cv_sprayoffsety = CV_RegisterVar({
	name = "sprayoffset_y",
	defaultvalue = 0
})

-- random chance that someone gets sprays each round
local cv_spraychance = CV_RegisterVar({
	name = "spray_chance",
	defaultvalue = "0.05",
	flags = CV_NETVAR|CV_FLOAT,
	possiblevalue = { MIN = 0, MAX = FRACUNIT }
})

-- random chance for hell maps
local cv_sprayhellchance = CV_RegisterVar({
	name = "spray_hellchance",
	defaultvalue = "0.25",
	flags = CV_NETVAR|CV_FLOAT,
	possiblevalue = { MIN = 0, MAX = FRACUNIT }
})

-- number of sprays each player gets
local cv_spraycount = CV_RegisterVar({
	name = "spray_count",
	defaultvalue = 3,
	flags = CV_NETVAR,
	possiblevalue = CV_Natural
})

-- delay before players get sprays, +/- 3.6 seconds
-- measured from race start
local cv_spraydelay = CV_RegisterVar({
	name = "spray_delay",
	defaultvalue = 10,
	flags = CV_NETVAR,
	possiblevalue = CV_Unsigned
})

-- wheter sprays can stick on players
local cv_sprayonplayers = CV_RegisterVar({
	name = "spray_onplayers",
	defaultvalue = "On",
	flags = CV_NETVAR,
	possiblevalue = CV_OnOff
})

-- how big wall sprays are
local cv_sprayscale = CV_RegisterVar({
	name = "spray_scale",
	defaultvalue = "1.5",
	flags = CV_NETVAR|CV_FLOAT,
	possiblevalue = CV_Natural
})

-- how big player sprays are
local cv_sprayplayerscale = CV_RegisterVar({
	name = "spray_playerscale",
	defaultvalue = "0.75",
	flags = CV_NETVAR|CV_FLOAT,
	possiblevalue = CV_Natural
})

-- how long sprays stick to players
local cv_sprayplayertime = CV_RegisterVar({
	name = "spray_playertime",
	defaultvalue = 8,
	flags = CV_NETVAR,
	possiblevalue = CV_Natural
})

-- minimum number of players required
local cv_sprayminplayers = CV_RegisterVar({
	name = "spray_minimumplayers",
	defaultvalue = 2,
	flags = CV_NETVAR,
	possiblevalue = { MIN = 1, MAX = #players }
})
-- would've been spray_minplayers, but...
-- Variables spray_minplayers and nmr_nohellrounds have same netid

-- wheter to allow custom buttons for sprays. useful if gameplay mods are already using custom buttons
local cv_sprayusecustom = CV_RegisterVar({
	name = "spray_usecustom",
	defaultvalue = "On",
	flags = CV_NETVAR|CV_CALL|CV_NOINIT,
	possiblevalue = CV_OnOff,
	func = function() showtip = true end -- show tip again so people know custom buttons are no longer usable
})

-- force a specific spray
local cv_sprayforce = CV_RegisterVar({
	name = "spray_force",
	defaultvalue = "0",
	flags = CV_NETVAR|CV_CALL,
	possiblevalue = { MIN = 0, MAX = 1024*64 },
	func = function(cv)
		if consoleplayer == server and cv.value > framecount then
			CONS_Printf(server, "Too high, limit is "..framecount)
			CV_StealthSet(cv, 0)
		end
	end
})

-- make sprays bounce off walls multiple times
-- (and give everyone infinite sprays)
local cv_spraybounce = CV_RegisterVar({
	name = "spray_chaos",
	defaultvalue = "Off",
	flags = CV_NETVAR|CV_CALL|CV_NOINIT,
	possiblevalue = CV_OnOff,
	func = function()
		stickychanged = true
	end
})

local function initspray(p)
	p.spray = {}
	local s = p.spray
	s.spraycount = 0 -- number of sprays
	s.gettime = starttime + P_SignedRandom() + cv_spraydelay.value*TICRATE -- when they will get sprays
	s.spraycharge = 0 -- delay before getting sprays
	s.spraypressed = false -- pressed button?
	setmetatable(s, {
		__index = function(t, k) error(k..": no peeking", 2) return end,
		__newindex = function(t, k, v) error(k..": no touchy", 2) return end
	})
end

COM_AddCommand("givespray", function(p, target, count)
	local match, num = SG_FindPlayer(target)
	if target == nil then match, num = p, 1 end
	if count then count = tonumber($) else count = 1000000 end

	if match == nil then CONS_Printf(p, num == 0 and "Couldn't find player" or "Player number doesn't exist"); return end
	if num ~= 1 then CONS_Printf(p, "Too many players matched "..target); return end

	if not match.spray then initspray(match) end
	match.spray.gettime = -1
	match.spray.spraycount = max(0, $ + count)

	local text = count == 1000000 and "a lifetime supply of" or tostring(count)
	CONS_Printf(p, "Giving "..text.." sprays to "..match.name)
	CONS_Printf(match, "Sweet! You've been gifted "..text.." sprays")
	S_StartSound(nil, sfx_sprygt, match)
end, COM_ADMIN)

-- commands to use sprays, useful for binding extra keys
for i = 1, 4 do
	COM_AddCommand("usespray"..(i ~= 1 and tostring(i) or ""), function(p)
		usespray[i-1] = true
	end, COM_LOCAL)
end

-- hand out sprays to everyone
addHook("MapLoad", function(mapnum)
	-- count players first, and reset spray data
	local playercount = 0
	for p in players.iterate do
		p.spray = nil
		if not p.spectator then playercount = $ + 1 end
	end
	if playercount < cv_sprayminplayers.value then return end

	local chance = (mapheaderinfo[mapnum].menuflags & LF2_HIDEINMENU) and cv_sprayhellchance.value or cv_spraychance.value
	for p in players.iterate do
		if not p.spectator and chance > P_RandomFixed() then
			initspray(p)
		end
	end
end)

-- helper for local disable
local function playsound(source, sound, target)
	if cv_sprayenabled.value then S_StartSound(source, sound, target) end
end

-- this will most likely have dire consequences for this game's community
addHook("PlayerCmd", function(p, cmd)
	local i = p.splitscreenindex
	if cv_sprayenabled.value and (cv_sprayusecustom.value and (cv_spraybutton[i].value and ((cmd.buttons & cv_spraybutton[i].value) == cv_spraybutton[i].value)) or usespray[i]) then
		cmd.buttons = $ | BT_SPRAY
	end
	usespray[i] = false
end)

-- player stuff
addHook("PlayerThink", function(p)
	local s = p.spray
	if cv_spraybounce.value and (not s or s.gettime ~= -1) then
		initspray(p)
		s = p.spray
		s.gettime = -1
		s.spraycount = 1000000
	end
	if not (s and leveltime > 0) then return end
	-- spectators can only spray if they were manually given sprays
	-- ...actually why not?
	--if p.spectator and not (s.gettime == -1) then return end

	-- randomly chosen to get sprays?
	if s.gettime == leveltime then
		--print("Spray charge: "..p.name)
		s.spraycharge = CHARGE_TIME
		playsound(nil, CHARGESOUND, p)
	end

	-- charge up sprays
	if s.spraycharge > 0 then
		s.spraycharge = $ - 1
		if s.spraycharge == 0 then
			-- got it!
			s.spraycount = cv_spraycount.value
			playsound(nil, GOTSOUND, p)
			if cv_spraytip.value and showtip and not replayplayback then
				tiptimer = 6*TICRATE
				showtip = false
			end
		end
	end

	-- actually using the spray
	if (p.cmd.buttons & BT_SPRAY) and s.spraycount > 0 and not s.spraypressed then
		s.spraypressed = true

		-- access a spectator's mobj with this one simple trick!
		local oldspec = p.spectator
		p.spectator = false
		local pmo = p.mo
		p.spectator = oldspec

		local angleofs = FixedDiv(p.cmd.driftturn * 0x80000, max(FRACUNIT, p.speed/5))
		if p.spectator then angleofs = 0 end
		local spray = P_SPMAngle(pmo, MT_SPRAY, pmo.angle + angleofs)
		if not spray then
			-- can't spawn here apparently
			-- usually because ceiling is too low
			playsound(pmo, NOSPAWNSOUND)
			return
		end

		spray.sprite = SPR_GRFT
		spray.threshold = TICRATE/2
		spray.flags = $ | MF_BOUNCE | (cv_sprayonplayers.value * MF_SPECIAL)
		spray.scale = mapobjectscale / 16 -- don't use player's scale
		spray.destscale = mapobjectscale
		--spray.target = pmo -- done by P_SPMAngle
		if not cv_sprayenabled.value then spray.flags = $ | MF_NOSECTOR end

		local speed = p.speed + 20*mapobjectscale
		-- spectators get the privilege of vertical aiming
		if p.spectator then
			local limit = ANGLE_90 - ANG20
			local angle = max(-limit, min(p.aiming, limit))
			spray.momz = FixedMul(sin(angle), mapobjectscale*16)
			speed = FixedMul($, cos(angle))
			spray.z = $ - 16*FRACUNIT
		end

		P_InstaThrust(spray, pmo.angle + angleofs, speed)

		-- normally i wouldn't be against using variable names like "oldmomx",
		-- but if there are a lot, and i mean A LOOOOT of sprays in the map,
		-- it can adversely affect the size of the savegame
		-- so use hardcode variables as much as possible to minimize the impact on the savegame
		-- (and for a bit of speed)
		-- don't underestimate the power of 16 people hammering custom 2
		spray.extravalue1 = spray.momx
		spray.extravalue2 = spray.momy 

		s.spraycount = $ - 1
		playsound(pmo, LAUNCHSOUND)
		for dp in displayplayers.iterate do
			if p == dp then tiptimer = 0 end
		end
	elseif not (p.cmd.buttons & BT_SPRAY) then
		s.spraypressed = false
	end
end)

local function P_PointOnLineSide(x, y, line)
	local v1 = line.v1

	if not line.dx then
		if x <= v1.x then
			return line.dy > 0
		end
		return line.dy < 0
	end
	if not line.dy then
		if y <= v1.y then
			return line.dx < 0
		end
		return line.dx > 0
	end

	local dx = x - v1.x
	local dy = y - v1.y

	local left = FixedMul(line.dy/FRACUNIT, dx)
	local right = FixedMul(dy, line.dx/FRACUNIT)

	if right < left then
		return false // front side
	end
	return true // back side
end

-- the exact lineangle from bounce physics can be a bit flaky, so we need a LITTLE headroom
-- to avoid signed overflow bullshit, just mask off the low 24 bits (~1.4 degrees)
local function roundangle(angle)
	if angle >= 0 then
		return (angle + 0x800000) & ~0xFFFFFF
	else
		return (angle - 0x800000) & ~0xFFFFFF
	end
end

-- determines how sprays should stick to a sector
-- returns sector and (true if ceiling, else floor)
local function StickToSector(mo, sec)
	-- don't stick to thok barriers
	if sec.floorheight == sec.ceilingheight then
		return
	end

	-- stick to floor?
	if sec.floorheight > mo.z then
		return sec, false
	-- stick to ceiling?
	elseif sec.ceilingheight < mo.z + mo.height then 
		return sec, true
	-- check for FOFs
	else
		for ff in sec.ffloors() do
			if (ff.flags & (FF_EXISTS|FF_BLOCKOTHERS)) == (FF_EXISTS|FF_BLOCKOTHERS) -- don't stick to intangible FOFs
			and ff.bottomheight < mo.z + mo.height and ff.topheight > mo.z then
				-- TODO maybe pick floor or ceiling based on pegging flags?
				return ff.sector, false
			end
		end
	end
end

local function SetSprayFrame(mo, flags)
	local rand = P_RandomFixed() -- no RNG desyncs please
	local frame = cv_sprayforce.value or NotRandomRange(rand, 1, framecount)
	if isdedicatedserver then frame = 1 end
	mo.frame = flags | detectedframes[frame]:byte(5) - 0x41
	mo.sprite = _G["SPR_"..detectedframes[frame]:sub(1, 4)]
	mo.movecount = -cv_sprayforce.value or rand
end

-- spray thinker
addHook("MobjThinker", function(mo)
	-- wall spray (just skip the thinker)
	if mo.flags & MF_NOCLIP then
		if mo.scale ~= mo.destscale then
			mo.scalespeed = (7*(mo.destscale - mo.scale)/8) + FRACUNIT/1024
		elseif not (mo.movedir and not cv_spraybounce.value) then
			mo.flags = $ | MF_NOTHINK
			-- hyperoptimizing savegames
			mo.scalespeed = mapobjectscale/12
		end
		if mo.movedir and not cv_spraybounce.value then
			mo.z = sectors[mo.movedir][mo.threshold and "ceilingheight" or "floorheight"] + mo.movefactor
		end
		return true
	end

	-- stuck to player
	-- used to be a separate mobj, but i wanna save a freeslot
	if mo.flags & MF_SCENERY then
		local pmo = mo.tracer -- yes, tracer. target is the source player. it's confusing, but that's how every other projectile works
		if not pmo then
			mo.fuse = 1 -- player gone? destroy now
			return
		end

		P_MoveOrigin(mo, pmo.x, pmo.y, pmo.z)
		mo.destscale = FixedMul(pmo.scale, cv_sprayplayerscale.value)
		mo.scalespeed = FixedMul(pmo.scalespeed, cv_sprayplayerscale.value)
		K_MatchGenericExtraFlags(mo, pmo) -- resets MF2_DONTDRAW

		if mo.fuse <= TICRATE and pmo.player.kartstuff[k_hyudorotimer] == 0 then
			mo.flags2 = $ ^^ (leveltime & 1) * MF2_DONTDRAW
		end
		return
	end

	-- launched spray
	mo.threshold = max(0, $ - 1)

	-- wait for the spray to bounce off a wall
	if mo.momx == mo.extravalue1 and mo.momy == mo.extravalue2 then return end

	-- you might think that it would be easy to orient an mobj to the wall it collided with
	-- unfortunately, there is no such function available to my knowledge,
	-- and p_map.c's bestslideline isn't exposed anywhere,
	-- and this isn't 2.2 so there's no MobjLineCollide hook

	-- you know what we CAN do though?
	-- derive the angle of the line the spray collided with by abusing bounce physics
	local tmxmove = FixedMul(mo.extravalue1, (FRACUNIT - (FRACUNIT>>2) - (FRACUNIT>>3)))
	local tmymove = FixedMul(mo.extravalue2, (FRACUNIT - (FRACUNIT>>2) - (FRACUNIT>>3)))

	local linecos = tmxmove - mo.momx
	local linesin = tmymove - mo.momy
	local lineangle = R_PointToAngle2(0, 0, linecos, linesin) + ANGLE_90

	-- you know what would be even funnier?
	-- bringing back the blockmap search so we can stick to FOFs
	local blocklines = {}
	local ang1 = roundangle(lineangle)
	searchBlockmap("lines", function(_, foundline)
		local r = R_PointToAngle2(foundline.v1.x, foundline.v1.y, foundline.v2.x, foundline.v2.y)
		local ang2 = roundangle(r)
		local ang3 = roundangle(r + ANGLE_180)

		if ang1 == ang2 or ang1 == ang3 then
			table.insert(blocklines, foundline)
		end
	end, mo)

	-- sort by distance, just like old times
	local x, y = mo.x - mo.momx, mo.y - mo.momy
	table.sort(blocklines, function(a, b)
		return R_PointToDist2(x, y, P_ClosestPointOnLine(x, y, a)) < R_PointToDist2(x, y, P_ClosestPointOnLine(x, y, b))
	end)

	-- now go through all the lines and find a sector to stick to
	-- closest lines first!
	local stickysec, ceiling
	for i = #blocklines, 1, -1 do
		local bestline = blocklines[i]
		local target = P_PointOnLineSide(mo.x - mo.momx, mo.y - mo.momy, bestline)
		local sec = target and bestline.frontsector or bestline.backsector
		if sec then
			local sec, cel = StickToSector(mo, sec)
			if sec or not (bestline.flags & ML_TWOSIDED) then
				stickysec, ceiling = sec, cel
			end
		end
	end

	-- the only way to reset an mobj's angle interpolation is to spawn a new mobj
	local spray = P_SpawnMobj(mo.x - mo.momx, mo.y - mo.momy, mo.z, MT_SPRAY)
	SetSprayFrame(spray, FF_FULLBRIGHT | FF_PAPERSPRITE)
	if stickysec then
		spray.movedir = #stickysec
		spray.movefactor = mo.z - stickysec[ceiling and "ceilingheight" or "floorheight"]
		spray.threshold = ceiling and 1 or 0
	end

	spray.scale = mo.scale
	spray.destscale = FixedMul(cv_sprayscale.value, mapobjectscale)
	spray.scalespeed = 0

	K_MatchGenericExtraFlags(spray, mo)
	spray.flags = $ | MF_NOCLIP -- doesn't do anything, just a "is wall spray" flag
	spray.angle = lineangle
	if not cv_sprayenabled.value then spray.flags = $ | MF_NOSECTOR end

	P_SetOrigin(spray, spray.x, spray.y, spray.z) -- dumb hack to fix interpolation
	playsound(spray, SPRAYEDSOUND)

	if R_PointToDist2(0, 0, mo.momx, mo.momy) > mapobjectscale*10 and cv_spraybounce.value then
		mo.extravalue1 = mo.momx
		mo.extravalue2 = mo.momy
		mo.angle = lineangle + ANGLE_90
	else
		P_RemoveMobj(mo)
	end
end, MT_SPRAY)

-- spray has touched a player, time to spray 'em
addHook("TouchSpecial", function(mo, pmo)
	-- don't spray ourselves! unless the threshold is zero
	if mo.threshold > 0 and mo.target == pmo then return true end

	local spray = P_SpawnMobj(pmo.x, pmo.y, pmo.z, MT_SPRAY)
	SetSprayFrame(spray, FF_FULLBRIGHT)
	spray.scale = 9*FixedMul(pmo.scale, cv_sprayplayerscale.value)/10
	K_MatchGenericExtraFlags(spray, pmo)
	spray.flags = $ | MF_SCENERY
	spray.tracer = pmo
	spray.target = mo.target -- not necessary, but might as well
	spray.fuse = cv_sprayplayertime.value * TICRATE -- how long before it disintegrates
	if not cv_sprayenabled.value then spray.flags = $ | MF_NOSECTOR end

	playsound(spray, SPRAYEDSOUND)
	
	if cv_hl and cv_hl.value then
		local hit = {
			target = pmo.player,
			inflictor = mo,
			source = mo.target and mo.target.player,
			icon = "SPRAYCAN",
		}

		HL_AddHit(hit)
	end
end, MT_SPRAY)

-- spray disintegrating, after being stuck on a player
addHook("MobjFuse", function(mo)
	if mo.tracer then playsound(mo.tracer, DESTROYSOUND) end
	for i = 1, 5 do
		local dust = P_SpawnMobj(mo.x, mo.y, mo.z + FixedMul(40*FRACUNIT, mapobjectscale), MT_DRIFTDUST)
		if mo.tracer then dust.momx, dust.momy, dust.momz = mo.tracer.momx/2, mo.tracer.momy/2, mo.tracer.momz/2 end
		dust.momx = $ + FixedMul(P_SignedRandom() * FRACUNIT/16, mapobjectscale)
		dust.momy = $ + FixedMul(P_SignedRandom() * FRACUNIT/16, mapobjectscale)
		dust.momz = $ + FixedMul(P_SignedRandom(), mapobjectscale)
		if not cv_sprayenabled.value then dust.flags2 = $ | MF2_DONTDRAW end
	end
end, MT_SPRAY)

local buttons = {
	[0] = "", -- no button
	[BT_CUSTOM1] = "Press CUSTOM 1!",
	[BT_CUSTOM2] = "Press CUSTOM 2!",
	[BT_CUSTOM3] = "Press CUSTOM 3!",
	[BT_CUSTOM1|BT_CUSTOM2] = "Press CUSTOM 1+2!",
	[BT_CUSTOM2|BT_CUSTOM3] = "Press CUSTOM 2+3!",
	[BT_CUSTOM1|BT_CUSTOM3] = "Press CUSTOM 1+3!",
	[BT_ATTACK] = "Press ITEM!",
}

-- have to do this every time now in case of dedis
addHook("PlayerJoin", function(pnum)
	if not isserver and pnum == #consoleplayer then
		framecount = 0
		visibilitychanged = true
		stickychanged = true
	end
end)
addHook("ThinkFrame", function()
	if #detectedframes ~= framecount and #detectedframes > 1 then
		framecount = #detectedframes
		for mo in thinkers.iterate("mobj") do
			if mo.type == MT_SPRAY and mo.flags & (MF_NOCLIP|MF_SCENERY) then
				local frame = mo.movecount < 0 and -mo.movecount or NotRandomRange(mo.movecount, 1, framecount)
				mo.frame = (mo.frame & ~FF_FRAMEMASK) | detectedframes[frame]:byte(5) - 0x41
				mo.sprite = _G["SPR_"..detectedframes[frame]:sub(1, 4)]
			end
		end
	end
	if visibilitychanged then
		local off = cv_sprayenabled.value == 0
		for mo in thinkers.iterate("mobj") do
			if mo.type == MT_SPRAY then
				if off then
					mo.flags = $ | MF_NOSECTOR
				else
					mo.flags = $ & ~MF_NOSECTOR
				end
			end
		end
		visibilitychanged = false
	end
	if stickychanged then
		local stick = cv_spraybounce.value == 0 and cv_sprayenabled.value == 1
		for mo in thinkers.iterate("mobj") do
			if mo.type == MT_SPRAY and mo.flags & MF_NOCLIP then
				if stick and mo.movedir then
					mo.flags = $ & ~MF_NOTHINK
				elseif mo.scale == mo.destscale then
					mo.flags = $ | MF_NOTHINK
				end
			end
		end
		stickychanged = false
	end
end)

hud.add(function(v, p, c)
	-- autodetect the number of sprays
	if detectsprites then
		detectedframes = {}
		local c = 0x54 -- 'T'
		repeat
			local sprname = "GRF"..string.char(c)
			c = c + 1
			for i = (c == 0x55 and 0x42 or 0x41), 0x80 do -- B to ???
				if v.patchExists(sprname..string.char(i == 0x5c and 0x2b or i).."0") then
					table.insert(detectedframes, sprname..string.char(i))
				else
					break
				end
			end
		until detectedframes[#detectedframes]:byte(4) == c - 2

		print("\130SPRAYS: Found "..#detectedframes.." sprays")
		detectsprites = false
	end

	if not (cv_sprayenabled.value and p.spray) then return end

	local splitindex = c.pnum - 1

	local vflags = (SG_GetSplitscreenFlags and SG_GetSplitscreenFlags(splitindex) or 0) | V_HUDTRANS | V_SNAPTOLEFT | V_SNAPTOTOP
	local x, y = 42, 64
	local align = "center"
	local colormap = v.getColormap(TC_DEFAULT, p.skincolor)

	if splitscreen == 1 then
		x = 68
		y = 15
	elseif splitscreen >= 2 then
		-- 3P/4P splitscreen is text only
		vflags = $ | V_GREENMAP
		x = 30
		y = 4
		align = "left"
		if splitindex & 1 == 1 then
			-- flip to other side
			vflags = $ | V_SNAPTOLEFT ^^ V_SNAPTORIGHT
			x = 160 - $
			align = "right"
		end
	end

	x, y = x + cv_sprayoffsetx.value, y + cv_sprayoffsety.value
	if v.interpolate then v.interpolate(true) end

	-- has sprays?
	if p.spray.spraycount > 0 then
		if splitscreen <= 1 then v.draw(x - 12, y - 8, v.cachePatch("SPRAYCAN"), vflags, colormap) end
		v.drawString(x, y, (p.spray.spraycount > 999 and "MAX" or p.spray.spraycount), vflags, align)
		if tiptimer > 0 then
			local xofs, align = 28, "thin"
			if splitscreen > 1 and splitindex & 1 then
				align = "thin-right"
				xofs = -28
			end
			if splitscreen <= 1 then xofs = $/2 end

			-- make the text more readable, but fade it out
			vflags = $ & ~V_ALPHAMASK
			if tiptimer < TICRATE then vflags = $ | (10*(TICRATE - tiptimer)/TICRATE)*V_10TRANS end

			local text = cv_sprayusecustom.value and buttons[cv_spraybutton[splitindex].value] or 'Run cmd "usespray'..(splitindex ~= 0 and tostring(splitindex+1) or "")..'"!'
			v.drawString(x + xofs, y, text, vflags | V_GREENMAP | V_ALLOWLOWERCASE | V_6WIDTHSPACE, align)
			tiptimer = $ - 1
		end
	-- or is charging?
	elseif p.spray.spraycharge > 0 then
		if splitscreen <= 1 then
			v.draw(x - 12, y - 8 + sin(p.spray.spraycharge*ANGLE_45)/(FRACUNIT/4), v.cachePatch("SPRAYCAN"), vflags, colormap)
		else
			v.drawString(x, y, "...", vflags, align)
		end
	end
end)
